UNIT AcsFrm;

INTERFACE

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

Type
  TAcsForm = class (TForm)
    Memo: TMemo;
    RunButton: TButton;
    OpenDialog: TOpenDialog;
    procedure RunButtonClick (Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    procedure Send (txt: string);
    procedure Run (file_name: string);
  end;

Var
  AcsForm: TAcsForm;

IMPLEMENTATION

Uses AccCtrl2, AclApi2, TypInfo;

{$R *.DFM}

Function Select (n: integer; tab: array of string): string;
Begin
     if (n >= low (tab)) and (n <= high (tab))
        then result := tab [n]
        else result := '???';
End;

Function BoolToStr (b: boolean): string;
Begin
     if b then result := 'true'
          else result := 'false';
End;

Procedure TAcsForm.Send (txt: string);
Begin
     Memo.Lines.Append (txt);
End;

Function PSidToStr (psid: PSID): string;

Const use_table: array [1..8] of string =
      ( 'SidTypeUser',
        'SidTypeGroup',
        'SidTypeDomain',
        'SidTypeAlias',
        'SidTypeWellKnownGroup',
        'SidTypeDeletedAccount',
        'SidTypeInvalid',
        'SidTypeUnknown' );

Var  name_arr: array [0..255] of char;
     name_ptr: PChar;
     name_len: DWORD;

     domain_arr: array [0..255] of char;
     domain_ptr: PChar;
     domain_len: DWORD;

     use: SID_NAME_USE;
     ans: boolean;

Begin
     (*
     function LookupAccountSid (
        lpSystemName: PChar;
        Sid: PSID;
        Name: PChar;
        var cbName: DWORD;
        ReferencedDomainName: PChar;
        var cbReferencedDomainName: DWORD;
        var peUse: SID_NAME_USE): BOOL; stdcall;
     *)

     name_ptr := @ name_arr;
     name_len := sizeof (name_arr);

     domain_ptr := @ domain_arr;
     domain_len := sizeof (domain_arr);

     ans := LookupAccountSid (
        nil,	// address of string for system name
        psid,	// address of security identifier
        name_ptr,   // address of string for account name
        name_len,   // address of size account string
        domain_ptr, // address of string for referenced domain
        domain_len, // address of size domain string
        use); 	    // structure for SID type

     if ans then
        result := 'name=' + name_arr +
                  ', domain=' + domain_arr +
                  ', use=' + select (use-1, use_table)
     else
        result := 'error: ' + SysErrorMessage (GetLastError);

     result := '(' + result + ')';
End;

Procedure TAcsForm.Run (file_name: string);

const a = sizeof (TRUSTEE);
      b = sizeof (EXPLICIT_ACCESS);

Var  err: DWORD;
     owner: PSID;
     group: PSID;
     dacl: PACL;
     desc: PSECURITY_DESCRIPTOR;

     ans: BOOL;
     dacl2: PACL;
     daclPresent, daclDefaulted: BOOL;

     count: ULONG;
     list, ref: PEXPLICIT_ACCESS;
     elem: PEXPLICIT_ACCESS;
     i: integer;
     perm: DWORD;
     mode: ACCESS_MODE;
     txt: string;
     trustee: Trustee_A;
     sid: PSID;

Begin
     Memo.Lines.Clear;

     err := GetNamedSecurityInfo (

        PChar (file_name), // name of the object
        SE_FILE_OBJECT,	   // type of object

        1 * OWNER_SECURITY_INFORMATION or
        1 * GROUP_SECURITY_INFORMATION or
        1 * DACL_SECURITY_INFORMATION or
        0 * SACL_SECURITY_INFORMATION,     // type of security information to retrieve

        @ owner, // receives a pointer to the owner SID
        @ group, // receives a pointer to the primary group SID
        @ dacl,  // receives a pointer to the DACL
        nil,     // receives a pointer to the SACL
        desc);   // receives a pointer to the security descriptor

     send (file_name);
     if (err=ERROR_SUCCESS) then
        send ('GetNamedSecurityInfo O.K.')
     else begin
        send ('GetNamedSecurityInfo Error: ' + SysErrorMessage (err));
        exit;
     end;
     send ('');

     {--------------------------------------------------------------------}

     ans := GetSecurityDescriptorDacl (
        desc,           // address of security descriptor
        daclPresent,	// address of flag for presence of disc. ACL
        dacl2,          // address of pointer to ACL
        daclDefaulted); // address of flag for default disc. ACL
     Win32Check (ans);

     send ('dacl=dacl2   ' + BoolToStr (dacl=dacl2));
     send ('daclPresent   ' + BoolToStr (daclPresent));
     send ('daclDefaulted   ' + BoolToStr (daclDefaulted));
     send ('');

     {--------------------------------------------------------------------}

     send ('owner ' + PSidToStr (owner));
     send ('group ' + PSidToStr (group));
     send ('');

     {--------------------------------------------------------------------}

     (*
     function GetExplicitEntriesFromAcl (
        pacl: PACL;
        var pcCountOfExplicitEntries: ULONG;
        var pListOfExplicitEntries: PEXPLICIT_ACCESS_): DWORD; stdcall;
     *)

     if dacl = nil then begin
        send ('ACL is empty, may be not NTFS file system');
        exit;
     end;

     err := GetExplicitEntriesFromAcl (
        dacl,	// pointer to the ACL from which to get entries
        count,	// receives number of entries in the list
        list);  // receives pointer to list of entries

     if (err<>ERROR_SUCCESS) then begin
        send ('Error: ' + SysErrorMessage (err));
     end
     else begin
        send ('count = ' + IntToStr (count));
        send ('');

        elem := list;
        for i := 0 to count-1 do begin
           // elem := ref;
           // dword (ref) := dword (ref) + sizeof (elem^);

           perm := elem.grfAccessPermissions;
           send ('permisions ' + IntToHex (perm, 8));

           mode := elem.grfAccessMode;
           txt := GetEnumName (typeinfo (ACCESS_MODE), ord (mode));
           send ('access mode ' + txt);

           // elem.grfInheritance;

           trustee := elem.Trustee;

           txt := GetEnumName (typeinfo (TRUSTEE_FORM), ord (trustee.TrusteeForm));
           send ('trustee form ' + txt);

           txt := GetEnumName (typeinfo (TRUSTEE_TYPE), ord (trustee.TrusteeType));
           send ('trustee type ' + txt);

           if trustee.TrusteeForm = TRUSTEE_IS_NAME then begin
              if trustee.ptstrName <> nil then
                 send ('trustee name ' + string (trustee.ptstrName));
           end
           else if trustee.TrusteeForm = TRUSTEE_IS_SID then begin
              sid := trustee.ptstrName;
              if sid <> nil then
                 send ('trustee sid ' + PSidToStr (sid));
           end;

           send ('');
           inc (elem);
        end;


        Win32Check (LocalFree (longword (list)) = 0);
     end;
     send ('');

     {--------------------------------------------------------------------}

     Win32Check (LocalFree (longword (desc)) = 0);

End;

procedure TAcsForm.FormCreate(Sender: TObject);
begin
     // RunButtonClick (sender);
end;

Procedure TAcsForm.RunButtonClick (Sender: TObject);
Begin
     if OpenDialog.Execute then
        Run (OpenDialog.FileName);
End;

END.
