UNIT SemPro;

INTERFACE

Uses Windows, {CreateSemaphore}
     SysUtils; {RaiseLastWin32Error}

Const DataSize = 1000;

Type TElement = record
        num: integer;
     end;

     PBlock = ^ TBlock;

     TBlock = record
        prev: PBlock;    // predchazejici blok
        next: PBlock;    // nasledujici blok

        data: array [1..DataSize] of TElement;
     end;

     TQueue = class
     private
        store_blk: PBlock;    // zapisovany blok
        store_inx: integer;   // pozice v bloku
        store_cnt: integer;   // pocet zapsanych prvku
        store_lim: integer;   // pocet neoznamenych prvku

        recall_blk: PBlock;   // cteny blok
        recall_inx: integer;  // pozice v bloku
        recall_cnt: integer;  // pocet prectenych prvku
        recall_lim: integer;  // pocet rezervovanych prvku
        recall_sem: THandle;  // semafor

        empty_count: integer;             // pocet prazdnych bloku
        empty_lock:  TRtlCriticalSection; // pristup k prazdnym blokum

        block_count: integer; // pocet alokovanych bloku

        procedure NewStoreBlock;

     public
        constructor Create;
        destructor Destroy; override;
        function BlockCount: integer;
        procedure Store (const elem: TElement);
        procedure Recall (out elem: TElement; out ready: boolean);
     end;

IMPLEMENTATION

(********************************* QUEUE **********************************)

Const group = 1;

Function NewBlock: PBlock;
Begin
     new (result);
End;

Constructor TQueue.Create;
Const sem_max = 1000*1000*1000;
Var   blk: PBlock;
Begin
     blk := NewBlock; // vytvor prvni blok

     blk.prev := blk; // cyklus obsahujici jeden blok
     blk.next := blk;

     store_blk := blk; // uklada se do prvniho bloku
     store_inx := 0;
     store_cnt := 0;
     store_lim := 0;

     recall_blk := blk; // cte se z prvniho bloku
     recall_inx := 0;
     recall_cnt := 0;
     recall_lim := 0;

     recall_sem := CreateSemaphore (nil, 0, sem_max, nil);
     if recall_sem = 0 then RaiseLastWin32Error;

     empty_count := 0;
     InitializeCriticalSection (empty_lock);

     block_count := 1;

     { vzdy mam nejaky blok => recall_blk <> nil, store_blk <> nil }
End;

Destructor TQueue.Destroy;
Var  p, t: PBlock;
Begin
     p := store_blk;
     while p <> store_blk do begin
        t := p;
        p := p.next;
        Dispose (t);
     end;

     store_blk := nil;
     store_inx := 0;

     recall_blk := nil;
     recall_inx := 0;

     CloseHandle (recall_sem);
     DeleteCriticalSection (empty_lock);
End;

Function TQueue.BlockCount: integer;
Begin
     result := block_count;
End;

(********************************* RECALL *********************************)

Var dbg: integer;

Procedure TQueue.Recall (out elem: TElement; out ready: boolean);
Begin
     if group = 1 then
        WaitForSingleObject (recall_sem, INFINITE)
     else begin
        if recall_lim = 0 then begin
           WaitForSingleObject (recall_sem, INFINITE);
           recall_lim := group;
        end;
        dec (recall_lim);
     end;

     if recall_inx = DataSize then begin
        recall_inx := 0;
        recall_blk := recall_blk.next;

        EnterCriticalSection (empty_lock);
        inc (empty_count);
        LeaveCriticalSection (empty_lock);
     end;

     inc (recall_cnt);

     inc (recall_inx);
     Assert (recall_inx >= 1);
     Assert (recall_inx <= DataSize);
     elem := recall_blk.data [recall_inx];

     inc (recall_cnt);
     ready := true;
End;

(********************************* STORE **********************************)

Procedure TQueue.NewStoreBlock;
Var  blk: PBlock;
Begin
     blk := NewBlock;

     Assert (store_blk <> nil);
     Assert (store_blk.next <> nil);
     Assert (block_count >= 1);

     blk.prev := store_blk;
     blk.next := store_blk.next;

     blk.prev.next := blk;
     blk.next.prev := blk;

     inc (block_count);
End;

Procedure TQueue.Store (const elem: TElement);
Var  add: boolean;
Begin
     if store_inx = DataSize then begin
        // store_blk je plny

        add := true; // true => pridat novy blok

        EnterCriticalSection (empty_lock);
        if empty_count > 0 then begin
           // je pripraven prazdny blok
           add := false;
           dec (empty_count);
        end;
        LeaveCriticalSection (empty_lock);

        if add then NewStoreBlock;

        // posun na novy blok
        store_blk := store_blk.next;
        store_inx := 0;
     end;

     inc (store_cnt);

     inc (store_inx);
     Assert (store_inx >=1);
     Assert (store_inx <= DataSize);
     store_blk^.data [store_inx] := elem;

     if group = 1 then
        ReleaseSemaphore (recall_sem, 1, nil)
     else begin
        inc (store_lim);
        if store_lim = group then begin
           ReleaseSemaphore (recall_sem, 1, nil);
           store_lim := 0;
        end;
     end;

     (*
     if store_cnt = 1000 then begin
        QueueFrm.receive := false;
        dbg := 1;
     end;
     *)
End;

END.
